import numpy as np
from autograd import numpy as anp
from autograd import grad
import joblib
import h5py
from scipy.signal import fftconvolve
from scipy.signal import find_peaks_cwt
from scipy.optimize import fmin_l_bfgs_b
from sklearn.decomposition import PCA, IncrementalPCA
from sklearn.model_selection import train_test_split

import logging

import matplotlib.pyplot as plt

from typing import Any, Dict, List, Optional

def matching_ids(a: np.ndarray, b: np.ndarray, c: np.ndarray) -> np.ndarray:
    """Returns list of train IDs common to sets a, b and c."""
    unique_ids = list(set(a).intersection(b).intersection(c))
    return np.array(unique_ids)

class PromptNotFoundError(Exception):
    """
    Exception representing the error condition generated by not finding the prompt peak.
    """
    def __init__(self):
        pass
    def __str__(self) -> str:
        return "No prompt peak has been detected."

class Model(object):
    """
    Object representing a previous fit of the model to be used to predict high-resolution
    spectrum from a low-resolution one.

    Args:
      channels: Selected channels to use as an input for the low resolution data.
      n_pca_lr: Number of low-resolution data PCA components.
      n_pca_hr: Number of high-resolution data PCA components.
      high_res_sigma: Resolution of the high-resolution spectrometer in electron-Volts.
      tof_start: Start looking at this index from the low-resolution spectrometer data. Set to None to perform no selection
      delta_tof: Number of components to take from the low-resolution spectrometer. Set to None to perform no selection.
      validation_size: Fraction (number between 0 and 1) of the data to take for validation and systematic uncertainty estimate.

    """
    def __init__(self,
                 channels:List[str]=["channel_1_D",
                                     "channel_2_B",
                                     "channel_3_A",
                                     "channel_3_B",
                                     "channel_4_C",
                                     "channel_4_D"],
                 n_pca_lr: int=400,
                 n_pca_hr: int=20,
                 high_res_sigma: float=0.2,
                 tof_start: Optional[int]=31445,
                 delta_tof: Optional[int]=200,
                 validation_size: float=0.05):
        self.channels = channels
        self.n_pca_lr = n_pca_lr
        self.n_pca_hr = n_pca_hr

        # PCA models
        self.lr_pca = IncrementalPCA(n_pca_lr, whiten=True)
        self.hr_pca = PCA(n_pca_hr, whiten=True)

        # PCA unc. in high resolution
        self.high_pca_unc: Optional[np.ndarray] = None

        # fit model
        self.fit_model = FitModel()

        # size of the test subset
        self.validation_size = validation_size

        # where to cut on the ToF PES data
        self.tof_start = tof_start
        self.delta_tof = delta_tof

        # high-resolution photon energy axis
        self.high_res_photon_energy: Optional[np.ndarray] = None

        # smoothing of the SPEC data in eV
        self.high_res_sigma = high_res_sigma

    def preprocess_low_res(self, low_res_data: Dict[str, np.ndarray]) -> np.ndarray:
        """
        Get a dictionary with the channel names for the inut low resolution data and output
        only the relevant input data in an array.

        Args:
          low_res_data: Dictionary with keys named channel_{i}_{k}, where i is a number between 1 and 4 and k is a letter between A and D.

        Returns: Concatenated and pre-processed low-resolution data of shape (train_id, features).
        """
        items = [low_res_data[k] for k in self.channels]
        if self.delta_tof is not None:
            items = [item[:, self.tof_start:(self.tof_start + self.delta_tof)] for item in items]
        else:
            items = [item[:, self.tof_start:] for item in items]
        cat = np.concatenate(items, axis=1)
        return cat

    def preprocess_high_res(self, high_res_data: np.ndarray, high_res_photon_energy: np.ndarray) -> np.ndarray:
        """
        Get the high resolution data and preprocess it.

        Args:
          high_res_data: High resolution data with shape (train_id, features).
          high_res_photon_energy: High resolution photon energy values (the "x"-axis of the high resolution data) with shape (train_id, features).

        Returns: Pre-processed high-resolution data of shape (train_id, features) before.
        """
        # Apply smoothing
        n_features = high_res_data.shape[1]
        mu = high_res_photon_energy[:, n_features//2, np.newaxis]
        gaussian = np.exp(-0.5*(high_res_photon_energy - mu)**2/self.high_res_sigma**2)
        gaussian /= np.sum(gaussian, axis=1, keepdims=True)
        high_res_gc = fftconvolve(high_res_data, gaussian, mode="same", axes=1)
        return high_res_gc

    def estimate_prompt_peak(self, low_res_data: Dict[str, np.ndarray]) -> int:
        """
        Estimate the prompt peak index.

        Args:
          low_res_data: Low resolution data with a dictionary containing the channel names.

        Returns: The prompt peak index.
        """
        # reduce on channel and on train ID
        sum_low_res = - np.mean(sum(list(low_res_data.values())), axis=0)
        widths = np.arange(10, 50, step=5)
        peak_idx = find_peaks_cwt(sum_low_res, widths)
        if len(peak_idx) < 1:
            raise PromptNotFoundError()
        peak_idx = sorted(peak_idx, key=lambda k: np.fabs(sum_low_res[k]), reverse=True)
        return peak_idx[0]

    def debug_peak_finding(self, low_res_data: Dict[str, np.ndarray], filename: str):
        """
        Produce image to understand if the peak finding step worked well.

        Args:
          low_res_data: Low resolution data with a dictionary containing the channel names.
          filename: The file name where to save the plot.

        """
        sum_low_res = - np.mean(sum(list(low_res_data.values())), axis=0)
        peak_idx = self.estimate_prompt_peak(low_res_data)
        fig = plt.figure(figsize=(8, 16))
        ax = plt.gca()
        ax.plot(np.arange(peak_idx-100, peak_idx+300),
                sum_low_res[peak_idx-100:peak_idx+300],
                c="b",
                label="Data")
        ax.set(title="",
               xlabel="Photon Spectrometer channel",
               ylabel="Sum of all Photon Spectrometer channels")
        plt.axvline(100,
                linewidth=3,
                ls="--",
                color='r',
                label="Peak position")
        ax.legend()
        plt.savefig(filename)
        plt.close(fig)

    def fit(self, low_res_data: Dict[str, np.ndarray], high_res_data: np.ndarray, high_res_photon_energy: np.ndarray) -> np.ndarray:
        """
        Train the model.

        Args:
          low_res_data: Low resolution data as a dictionary with the key set to `channel_{i}_{k}`, where i is a number between 1 and 4 and k is a letter between A and D. For each dictionary entry, a numpy array is expected with shape (train_id, ToF channel).
          high_res_data: Reference high resolution data with a one-to-one match to the low resolution data in the train_id dimension. Shape (train_id, ToF channel).
          high_res_photon_energy: Photon energy axis for the high-resolution data.

        Returns: Smoothened high resolution spectrum.
        """

        self.high_res_photon_energy = high_res_photon_energy

        print("Find peaks.")
        # if the prompt peak has not been given, guess it
        if self.tof_start is None:
            self.tof_start = self.estimate_prompt_peak(low_res_data)
            print("Prompt at", self.tof_start)

        print("Pre-processing low")
        low_res = self.preprocess_low_res(low_res_data)
        print("Pre-processing high")
        high_res = self.preprocess_high_res(high_res_data, high_res_photon_energy)
        # fit PCA
        print("PCA low", low_res.shape)
        low_pca = self.lr_pca.fit_transform(low_res)
        print("PCA high")
        high_pca = self.hr_pca.fit_transform(high_res)
        print("Split")
        # split in train and test for PCA uncertainty evaluation
        low_pca_train, low_pca_test, high_pca_train, high_pca_test = train_test_split(low_pca, high_pca, test_size=self.validation_size, random_state=42)
        # fit the linear model
        print("Fit")
        self.fit_model.fit(low_pca_train, high_pca_train, low_pca_test, high_pca_test)

        print("PCA unc")
        high_pca_rec = self.hr_pca.inverse_transform(high_pca)
        self.high_pca_unc =  np.sqrt(np.mean((high_res - high_pca_rec)**2, axis=0, keepdims=True))

        print("Done")

        return high_res

    def predict(self, low_res_data: Dict[str, np.ndarray]) -> np.ndarray:
        """
        Predict a high-resolution spectrum from a low resolution given one.
        The output includes the uncertainty in its second and third entries of the first dimension.

        Args:
          low_res_data: Low resolution data as in the fit step with shape (train_id, channel, ToF channel).

        Returns: High resolution data with shape (train_id, ToF channel, 3). The component 0 of the last dimension is the predicted spectrum. Components 1 and 2 correspond to two sources of uncertainty.
        """
        low_res = self.preprocess_low_res(low_res_data)
        low_pca = self.lr_pca.transform(low_res)
        n_trains = low_res.shape[0]
        # Get high res.
        high_pca = self.fit_model.predict(low_pca)
        high_res_predicted = self.hr_pca.inverse_transform(high_pca["Y"])
        n_high_res_features = high_res_predicted.shape[1]
        high_res_unc = self.hr_pca.inverse_transform(high_pca["Y"] + high_pca["Y_eps"]) - high_res_predicted
        result = np.stack((high_res_predicted, high_res_unc, np.broadcast_to(self.high_pca_unc, (n_trains, n_high_res_features))), axis=2)
        return result

    def save(self, filename: str, lr_pca_filename: str, hr_pca_filename: str):
        """
        Save the fit model in a file.

        Args:
          filename: H5 file name where to save this.
          lr_pca_filename: Name of the file where to save the low-resolution PCA decomposition.
          hr_pca_filename: Name of the file where to save the high-resolution PCA decomposition.
        """
        with h5py.File(filename, 'w') as hf:
            d = self.fit_model.as_dict()
            for key, value in d.items():
                if isinstance(value, int):
                    hf.attrs[key] = value
                else:
                    hf.create_dataset(key, data=value)
        joblib.dump(self.lr_pca, lr_pca_filename)
        joblib.dump(self.hr_pca, hr_pca_filename)


    def load(self, filename: str, lr_pca_filename: str, hr_pca_filename: str):
        """
        Load model from a file.

        Args:
          filename: Name of the file where to read the model from.
          lr_pca_filename: Name of the file from where to load the low-resolution PCA decomposition.
          hr_pca_filename: Name of the file from where to load the high-resolution PCA decomposition.

        """
        with h5py.File(filename, 'r') as hf:
            d = {k: hf[k][()] for k in hf.keys()}
            d.update({k: hf.attrs[k] for k in hf.attrs})
            self.fit_model.from_dict(d)
        self.lr_pca = joblib.load(lr_pca_filename)
        self.hr_pca = joblib.load(hr_pca_filename)

class FitModel(object):
    """
    Linear regression model with uncertainties.
    """
    def __init__(self):
        # training dataset
        self.X_train: Optional[np.ndarray] = None
        self.Y_train: Optional[np.ndarray] = None

        # test dataset to evaluate uncertainty
        self.X_test: Optional[np.ndarray] = None
        self.Y_test: Optional[np.ndarray] = None

        # normalized target
        self.Y_train_norm = None
        self.Y_test_norm = None

        # model parameter sizes
        self.Nx: int = 0
        self.Ny: int = 0

        # fit result
        self.A_inf: np.ndarray = None
        self.b_inf: np.ndarray = None
        self.u_inf: np.ndarray = None

        # fit monitoring
        self.loss_train: List[float] = list()
        self.loss_test: List[float] = list()

        self.input_data = None

    def fit(self, X_train: np.ndarray, Y_train: np.ndarray, X_test: np.ndarray, Y_test: np.ndarray):
        """
        Perform the fit and evaluate uncertainties with the test set.
        """

        # training dataset
        self.X_train: np.ndarray = X_train
        self.Y_train: np.ndarray = Y_train

        # test dataset to evaluate uncertainty
        self.X_test: np.ndarray = X_test
        self.Y_test: np.ndarray = Y_test

        # model parameter sizes
        self.Nx: int = self.X_train.shape[1]
        self.Ny: int = self.Y_train.shape[1]

        # initial parameter values
        A0: np.ndarray = np.eye(self.Nx, self.Ny).reshape(self.Nx*self.Ny)
        Aeps: np.ndarray = np.zeros(self.Nx)
        b0: np.ndarray = np.zeros(self.Ny)
        u0: np.ndarray = np.zeros(self.Ny)
        x0: np.ndarray = np.concatenate((A0, b0, u0, Aeps), axis=0)

        # reset loss monitoring
        self.loss_train: List[float] = list()
        self.loss_test: List[float] = list()

        def loss(x: np.ndarray, X: np.ndarray, Y: np.ndarray) -> float:
            """
            Calculate the loss function value for a given parameter set `x`.

            Args:
              x: The parameters as a flat array.
              X: The independent-variable dataset.
              Y: The dependent-variable dataset.

            Returns: The loss.
            """
            # diag( (in @ x - out) @ (in @ x - out)^T )
            A = x[:self.Nx*self.Ny].reshape((self.Nx, self.Ny))
            b = x[self.Nx*self.Ny:(self.Nx*self.Ny+self.Ny)].reshape((1, self.Ny))

            b_eps = x[(self.Nx*self.Ny+self.Ny):(self.Nx*self.Ny+self.Ny+self.Ny)].reshape((1, self.Ny))
            A_eps = x[(self.Nx*self.Ny+self.Ny+self.Ny):].reshape((self.Nx, 1))
            log_unc = X @ A_eps + b_eps

            #log_unc = anp.log(anp.exp(log_unc) + anp.exp(log_eps))
            iunc2 = anp.exp(-2*log_unc)

            #print("iunc2", iunc2)
            #print("log_unc", log_unc)

            return anp.mean( (0.5*((X@ A + b - Y)**2)*iunc2 + log_unc).sum(axis=1), axis=0 ) # Put RELU on (XX@x) and introduce new matrix W

        def loss_history(x: np.ndarray) -> float:
            """
            Calculate the loss function and keep a history of it in training and testing.

            Args:
              x: Parameters flattened out.

            Returns: The loss value.
            """
            l_train = loss(x, self.X_train, self.Y_train)
            l_test = loss(x, self.X_test, self.Y_test)

            self.loss_train += [l_train]
            self.loss_test += [l_test]
            return l_train

        def loss_train(x: np.ndarray) -> float:
            """
            Calculate the loss function for the training dataset.

            Args:
              x: Parameters flattened out.

            Returns: The loss value.
            """
            l_train = loss(x, self.X_train, self.Y_train)
            return l_train

        grad_loss = grad(loss_train)
        sc_op = fmin_l_bfgs_b(loss_history,
                              x0,
                              grad_loss,
                              disp=True,
                              factr=1e7,
                              maxiter=100,
                              iprint=0)

        # Inference
        self.A_inf = sc_op[0][:self.Nx*self.Ny].reshape(self.Nx, self.Ny)
        self.b_inf = sc_op[0][self.Nx*self.Ny:(self.Nx*self.Ny+self.Ny)].reshape(1, self.Ny)
        self.u_inf = sc_op[0][(self.Nx*self.Ny+self.Ny):(self.Nx*self.Ny+self.Ny+self.Ny)].reshape(1, self.Ny) # removed np.exp
        self.A_eps = sc_op[0][(self.Nx*self.Ny+self.Ny+self.Ny):].reshape(self.Nx, 1)

    def as_dict(self) -> Dict[str, Any]:
        """
        Export relevant variables to rebuild this object from a simple dictionary.

        Args:

        Returns: Dictionary with all relevant variables.
        """
        return dict(
                    X_train=self.X_train,
                    X_test=self.X_test,
                    Y_train=self.Y_train,
                    Y_test=self.Y_test,
                    A_inf=self.A_inf,
                    b_inf=self.b_inf,
                    u_inf=self.u_inf,
                    A_eps=self.A_eps,
                    loss_train=self.loss_train,
                    loss_test=self.loss_test,
                    Nx=self.Nx,
                    Ny=self.Ny)
    def from_dict(self, in_dict: Dict[str, Any]):
        """
        Rebuild this object from a dictionary.

        Args:
          in_dict: The input dictionary with relevant variables.

        """
        self.X_train = in_dict["X_train"]
        self.X_test = in_dict["X_test"]
        self.Y_train = in_dict["Y_train"]
        self.Y_test = in_dict["Y_test"]
        self.A_inf = in_dict["A_inf"]
        self.b_inf = in_dict["b_inf"]
        self.u_inf = in_dict["u_inf"]
        self.A_eps = in_dict["A_eps"]
        self.loss_train = in_dict["loss_train"]
        self.loss_test = in_dict["loss_test"]
        self.Nx = in_dict["Nx"]
        self.Ny = in_dict["Ny"]

    def predict(self, X: np.ndarray) -> Dict[str, np.ndarray]:
        """
        Predict y from X.

        Args:
          X: Input dataset.

        Returns: Predicted Y.
        """

        result = dict()

        # result
        result["Y"] = (X @ self.A_inf + self.b_inf)
        # flat uncertainty
        result["Y_unc"] = self.u_inf[0,:]
        # input-dependent uncertainty
        result["Y_eps"] = np.exp(X @ self.A_eps + result["Y_unc"])

        #self.result["res"] = self.model["spec_pca_model"].inverse_transform(self.result["res_pca"]) # transform PCA space to real space
        #self.result["res_unc"] = self.model["spec_pca_model"].inverse_transform(self.result["res_pca"] + self.model["u_inf"]) - self.model["spec_pca_model"].inverse_transform(self.result["res_pca"] ) 
        #self.result["res_unc"] = np.fabs(self.result["res_unc"])
        #self.result["res_eps"] = self.model["spec_pca_model"].inverse_transform(self.result["res_pca"] + self.result["res_pca_eps"]) - self.model["spec_pca_model"].inverse_transform(self.result["res_pca"] ) 
        #self.result["res_eps"] = np.fabs(self.result["res_eps"])
        #self.Yhat_pca = self.model["spec_pca_model"].inverse_transform(self.model["Y_test"])
        #self.result["res_unc_specpca"] =  np.sqrt(((self.Yhat_pca - self.model["spec_target"])**2).mean(axis=0))
        #self.result["res_unc_total"] =  np.sqrt(self.result["res_eps"]**2 + self.result["res_unc_specpca"]**2)
        return result

